Package: R6
Title: Encapsulated Classes with Reference Semantics
Version: 2.6.1
Authors@R: c(
    person("Winston", "Chang", , "winston@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Creates classes with reference semantics, similar to R's
    built-in reference classes. Compared to reference classes, R6 classes
    are simpler and lighter-weight, and they are not built on S4 classes
    so they do not require the methods package. These classes allow public
    and private members, and they support inheritance, even when the
    classes are defined in different packages.
License: MIT + file LICENSE
URL: https://r6.r-lib.org, https://github.com/r-lib/R6
BugReports: https://github.com/r-lib/R6/issues
Depends: R (>= 3.6)
Suggests: lobstr, testthat (>= 3.0.0)
Config/Needs/website: tidyverse/tidytemplate, ggplot2, microbenchmark,
        scales
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2025-02-14 21:15:19 UTC; winston
Author: Winston Chang [aut, cre],
  Posit Software, PBC [cph, fnd]
Maintainer: Winston Chang <winston@posit.co>
Repository: CRAN
Date/Publication: 2025-02-15 00:50:02 UTC
Built: R 4.6.0; ; 2026-04-25 15:39:09 UTC; unix
